#!/usr/bin/env python3
# encoding: utf-8
'''

check the format of node and param.

'''
from datetime import datetime
from ursgal import uparams
from ursgal import UController
from ursgal import ukb

# node format
node_meta_info = {
    'distributable' : {
        'essential' : False,
        'types'     : [bool],
    },
    'citation' : {
        'essential' : True,
        'types'     : [str],
    },
    # 'compress_raw_search_results' : {
    #     'essential' : False,
    #     'types'     : [bool],
    # },
    # 'cpu_usage' : {
    #     'essential' : False,
    #     'types'     : [int],
    # },
    'create_own_folder' : {
        'essential' : False,
        'types'     : [bool],
    },
    'edit_version' : {
        'essential' : True,
        'types'     : [float],
    },
    'engine' : {
        'essential' : False,
        'types'     : [dict],
    },
    # 'engine_exe' : {
    #     'essential' : False,
    #     'types'     : [dict],
    # },
    'engine_type' : {
        'essential' : True,
        'types'     : [dict],
    },
    # 'engine_url' : {
    #     'essential' : False,
    #     'types'     : [dict],
    # },
    'group_psms' : {
        'essential' : False,
        'types'     : [bool],
    },
    'in_development' : {
        'essential' : True,
        'types'     : [bool],
    },
    'include_in_git' : {
        'essential' : True,
        'types'     : [bool, type(None)],
    },
    'input_extensions' : {
        'essential' : True,
        'types'     : [list],
    },
    # 'input_multi_file' : {
    #     'essential' : True,
    #     'types'     : [bool],
    # },
    'mods_to_unimod_correction' : {
        'essential' : False,
        'types'     : [dict],
    },
    'name' : {
        'essential' : True,
        'types'     : [str],
    },
    'output_extensions' : {
        'essential' : True,
        'types'     : [list],
    },
    'output_suffix' : {
        'essential' : False,
        'types'     : [str, type(None)],
    },
    'rejected_output_suffix' : {
        'essential' : False,
        'types'     : [str],
    },
    'release_date' : {
        'essential' : True,
        'types'     : ['datetime', type(None)],
    },
    'utranslation_style' : {
        'essential' : True,
        'types'     : [str],
    },
    'version' : {
        'essential' : True,
        'types'     : [str],
    },
    'uses_unode' : {
        'essential' : False,
        'types'     : [str],
    },
}

# param format
param_info = {
    'available_in_unode' : {
        'essential'   : True,
        'types'       : [list],
        'ignore_key'  : ['ucontroller']
    },
    'default_value'   : {
        'essential'   : True,
        'types'       : [type(None), str, int, float, bool, list, tuple, dict],
    },
    'description' : {
        'essential'   : True,
        'types'       : [str],
    },
    'edit_version' : {
        'essential'   : True,
        'types'       : [float],
    },
    'triggers_rerun' : {
        'essential'   : True,
        'types'       : [bool],
    },
    'ukey_translation' : {
        'essential'   : True,
        'types'       : [dict],
    },
    'utag' : {
        'essential'   : True,
        'types'       : [list],
    },
    'uvalue_option'   : {
        'essential'   : True,
        'types'       : [dict],
    },
    'uvalue_translation' : {
        'essential'   : True,
        'types'       : [dict],
    },
    'uvalue_type' : {
        'essential'   : True,
        'types'       : [str]
    },
}

# type format
type_info = {
    'str' : {
        'my_type' : str,
        'option'  : {
            'multiple_line'  : [bool],
        },
    },
    'str_password' : {
        'my_type' : str,
        'option'  : {
            'multiple_line'  : [bool],
        },
    },
    'int' : {
        'my_type' : int,
        'option'  : {
            'max'            : [int],
            'min'            : [int],
            'updownval'      : [int],
            'unit'           : [str],
        },
    },
    'int _uevaluation_req' : {
        'my_type' : int,
        'option'  : {
            'max'            : [int, str],
            'min'            : [int, str],
            'updownval'      : [int],
            'unit'           : [str],
        },
    },
    'float' : {
        'my_types' : [int, float],
        'option'  : {
            'max'            : [int, float],
            'min'            : [int, float],
            'f-point'        : [float],
            'updownval'      : [int, float],
            'unit'           : [str],
        },
    },
    'bool' : {
        'my_type' : bool,
        'option'  : {
        },
    },
    'select' : {
        'my_type' : str,
        'option'  : {
            'select_type'       : [str],
            'available_values'  : [list],
            'custom_val_max'    : [int],
        },
    },
    'list' : {
        'my_type' : list,
        'option'  : {
            'item_title'     : [str],
            'item_type'      : [str],
            'custom_val_max' : [int],
        },
    },
    # 'tuple' : {
    #     'my_type' : tuple,
    #     'option': {
    #         'titles'     : [list],
    #         'type_dict'      : [dict],
    #         'custom_val_max' : [int],
    #     },
    # },
    'dict' : {
        'my_type' : dict,
        'option'  : {
            'item_titles'     : [dict],
            'value_types'      : [dict],
        },
    },
    'None' : {
        'my_type' : type(None),
        'option'  : {
            'system_param'   : [bool]
        },
    },
}

# styles which is required by system
style_list = [
    'ucontroller_style_1',
]

# pre-treatment of Node format
uc = UController()
all_node  = uc.unodes
all_node_name = list(all_node)
all_node_name.sort()
remove_list = []
for node_name in all_node_name:
    if node_name[0] == '_':
        if len(node_name) < 5 or node_name[0:5] != '_test':
            remove_list.append(node_name)
for remove_name in remove_list:
    all_node_name.remove(remove_name)

# pre-treatment of Param format
all_param = uparams.ursgal_params
all_param_name = list(all_param)
all_param_name.sort()
remove_list = []
for param_name in all_param_name:
    if param_name[0] == '_':
        if len(param_name) < 5 or param_name[0:5] != '_test':
            remove_list.append(param_name)
for remove_name in remove_list:
    all_param_name.remove(remove_name)


def chk_dict_keys( parent_name='', key_list=[], dict_item={} ):
    for k in key_list:
        if dict_item.get(k) is None and \
                dict_item.get(k, 'None') == 'None':
            parent_info = ''
            if parent_name != '':
                parent_info = ' in ' + str(parent_name)
            error_msg = str(k) + ' is required' + parent_info + '.'
            raise ValueError(error_msg)


def chk_vals_types( parent_name='', key_name='', vals=[], type_list=[] ):
    if len(type_list) == 0:
        return

    for v in vals:
        v_type = type(v)
        if ('datetime' in type_list) is True and v_type is str:
            num = len(v) - len(v.replace('-', '').replace(':', ''))
            if num == 2:
                try:
                    datetime.strptime(v, '%Y-%m-%d')
                    datetime_chk = True
                except:
                    datetime_chk = False
            elif num == 4:
                try:
                    datetime.strptime(v, '%Y-%m-%d %H:%M:%S')
                    datetime_chk = True
                except:
                    datetime_chk = False
        else:
            datetime_chk = False

        if (v_type in type_list) is not True and datetime_chk is not True:
            key_info = ''
            parent_info = ''
            if key_name != '':
                key_info = 'of \'' + str(key_name) + '\''
                if parent_name != '':
                    parent_info = ' in ' + str(parent_name) + ' '
            my_type = ''
            len_type_list = len(type_list)
            if len_type_list >= 1:
                num = 0
                for type_class in type_list:
                    if type_class == 'datetime':
                        type_name = 'datetime'
                    else:
                        type_name = type_class.__name__
                    if num == len_type_list-1:
                        if num == 0:
                            my_type = str(type_name)
                        else:
                            my_type = my_type + ' or ' + str(type_name)
                    elif num == 0:
                        my_type = str(type_name)
                    else:
                        my_type = my_type + ', ' + str(type_name)
                    num += 1
            error_msg = 'The type ' + key_info + parent_info + 'is not ' + \
                str(v_type.__name__) + '. ' + 'It has ' + my_type + '.'
            raise ValueError(error_msg)


def chk_json_item( parent_name='', essential=False, key_list=[], dict_item={}, \
        type_list=[] ):
    if essential is True:
        chk_dict_keys( parent_name, key_list, dict_item )

    for key_name in key_list:
        if dict_item.get(key_name) is not None or \
                dict_item.get(key_name) is None and \
                dict_item.get(key_name, 'None') != 'None':
            vals = [dict_item[key_name]]
            chk_vals_types( parent_name, key_name, vals, type_list )


def chk_format_node( node_name, node_dict ):
    print(node_name)
    chk_dict_keys(
        parent_name = node_name,
        key_list    = ['META_INFO'],
        dict_item   = node_dict,
    )
    chk_dict_keys(
        parent_name = 'node_meta_info',
        key_list    = list(node_dict['META_INFO']),
        dict_item   = node_meta_info,
    )

    for k, v in node_meta_info.items():
        chk_json_item(
            parent_name = 'META_INFO',
            essential   = v['essential'],
            key_list    = [k],
            dict_item   = node_dict['META_INFO'],
            type_list   = v['types'],
        )

    all_extensions    = list(ukb.FILE_EXTENSIONS)
    output_extensions = node_dict['META_INFO']['output_extensions']
    input_extensions  = node_dict['META_INFO']['input_extensions']
    for ext in (output_extensions + input_extensions):
        if (ext in all_extensions) is False:
            error_msg = '\'' + str(ext) + '\'' + ' could not be found in '\
                'uparams.ursgal_params[\'_extensions\'][\'default_value\'].keys()'
            raise ValueError(error_msg)

    global style_list
    style_list.append(node_dict['META_INFO']['utranslation_style'])


def search_dict_keys(dict_item, single_key=True, key_list=[], layer_num=0):
    layer_dict = {}
    for key in dict_item.keys():
        key_list.append(key)
        if isinstance(dict_item[key], dict):
            key_list2, layer_dict2 = search_dict_keys(
                dict_item[key],
                single_key,
                key_list,
                layer_num + 1,
            )
            for k, v in layer_dict2.items():
                if layer_dict.get(k) is None:
                    layer_dict[k] = []
                layer_dict[k].extend(v)

            if single_key is True:
                break
        else:
            if layer_dict.get(key) is None:
                layer_dict[key] = []
            layer_dict[key].append(layer_num+1)
    return key_list, layer_dict


def chk_format_param( param_name, param_dict ):
    chk_dict_keys(
        parent_name = 'param_info',
        key_list    = list(param_dict),
        dict_item   = param_info,
    )

    for k, v in param_info.items():
        chk_json_item(
            parent_name = param_name,
            essential   = v['essential'],
            key_list    = [k],
            dict_item   = param_dict,
            type_list   = v['types'],
        )

    for ava_node in param_dict['available_in_unode']:
        if (ava_node in param_info['available_in_unode']['ignore_key']) is False:
            chk_dict_keys(
                parent_name = 'wrappers',
                key_list    = [ava_node],
                dict_item   = all_node,
            )

    utag          = param_dict['utag']
    uvalue_option = param_dict['uvalue_option']
    uvalue_type   = param_dict['uvalue_type']
    default_value = param_dict['default_value']

    if ('input_files' in utag) is True:
        chk_json_item(
            parent_name = 'uvalue_option',
            essential   = True,
            key_list    = ['input_extensions'],
            dict_item   = uvalue_option,
            type_list   = [list],
        )

    global style_list
    style_list = list(set(style_list))
    ukey_translation_keys   = list(param_dict['ukey_translation'])
    uvalue_translation_keys = list(param_dict['uvalue_translation'])
    search_style_list = list(set(ukey_translation_keys+uvalue_translation_keys))

    for search_style in search_style_list:
        if (search_style in style_list) is False:
            error_msg = str(search_style) + ' is unknown style.'
            raise ValueError(error_msg)

    if (uvalue_type in list(type_info)) is False:
        error_msg = str(uvalue_type) + ' is unknown type.'
        raise ValueError(error_msg)

    if uvalue_type == 'str' or uvalue_type == 'str_password':
        my_type = type_info[uvalue_type]['my_type']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = [my_type, type(None)],
        )

        type_list = None
        if type(default_value) is type(None):
            type_list = [my_type]
        elif type(default_value) is not type(None):
            type_list = [my_type, type(None)]
        if type_list is not None:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['none_val'],
                dict_item   = uvalue_option,
                type_list   = type_list,
            )

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

    elif uvalue_type == 'int':
        my_type = type_info[uvalue_type]['my_type']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = [my_type, type(None)],
        )

        type_list = None
        if type(default_value) is type(None):
            type_list = [my_type]
        elif type(default_value) is not type(None):
            type_list = [my_type, type(None)]
        if type_list is not None:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['none_val'],
                dict_item   = uvalue_option,
                type_list   = type_list,
            )

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

    elif uvalue_type == 'int _uevaluation_req' and \
            type(default_value) is not type(None):
        my_type = type_info[uvalue_type]['my_type']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = [my_type, type(None)],
        )

        type_list = None
        if type(default_value) is type(None):
            type_list = [my_type]
        elif type(default_value) is not type(None):
            type_list = [my_type, type(None)]
        if type_list is not None:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['none_val'],
                dict_item   = uvalue_option,
                type_list   = type_list,
            )

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

    elif uvalue_type == 'float':
        my_types = type_info[uvalue_type]['my_types']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = my_types + [type(None)],
        )

        type_list = None
        if type(default_value) is type(None):
            type_list = my_types
        elif type(default_value) is not type(None):
            type_list = my_types + [type(None)]
        if type_list is not None:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['none_val'],
                dict_item   = uvalue_option,
                type_list   = type_list,
            )

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

    elif uvalue_type == 'bool':
        my_type = type_info[uvalue_type]['my_type']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = [my_type],
        )

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

    elif uvalue_type == 'select':
        my_type = type_info[uvalue_type]['my_type']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = [my_type],
        )

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

        chk_vals_types(
            parent_name = 'uvalue_option',
            key_name    = 'available_values',
            vals        = uvalue_option['available_values'],
            type_list   = [my_type],
        )

        chk_vals_types(
            parent_name = 'uvalue_option',
            key_name    = 'custom_val_max',
            vals        = [uvalue_option['custom_val_max']],
            type_list   = [int],
        )
        if uvalue_option['custom_val_max'] < 0:
            error_msg = 'custom_val_max is more than 0.'
            raise ValueError(error_msg)

    elif uvalue_type == 'list':
        my_type = type_info[uvalue_type]['my_type']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = [my_type, type(None)],
        )

        type_list = None
        if type(default_value) is type(None):
            type_list = [my_type]
        elif type(default_value) is not type(None):
            type_list = [my_type, type(None)]
        if type_list is not None:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['none_val'],
                dict_item   = uvalue_option,
                type_list   = type_list,
            )

        none_val   = uvalue_option['none_val']

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

        item_title = uvalue_option['item_title']
        item_type  = uvalue_option['item_type']

        # Why should checking for the same number of elements be necessary?
        # The number of elements in a list is variable
        # if default_value is not None:
        #     if len(title_list) != len(default_value):
        #         error_msg = 'default_value is not exactly number compared '\
        #             'with title_list.'
        #         raise ValueError(error_msg)
        # if none_val is not None:
        #     if len(title_list) != len(none_val):
        #         error_msg = 'none_val is not exactly number compared with '\
        #             'title_list.'
        #         raise ValueError(error_msg)

        # chk_json_item(
        #     parent_name = 'uvalue_option',
        #     essential   = True,
        #     key_list    = [item_title],
        #     dict_item   = item_type,
        #     type_list   = [str],
        # )

        # list_type_list = list(set(list(type_dict.values())))
        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )
        # for k, v in type_info.items():
        #     if k == item_type:
        #         chk_json_item(
        #             parent_name = 'uvalue_option',
        #             essential   = True,
        #             key_list    = list(v['option']),
        #             dict_item   = uvalue_option,
        #             type_list   = [eval(item_type)],
        #         )

        # Just one title for items in list now, so this should not be necessary anymore
        # for k, v in type_info.items():
        #     for title in title_list:
        #         if type_dict[title] == k:
        #             for k2, v2 in v['option'].items():
        #                 chk_json_item(
        #                     parent_name = k2,
        #                     essential   = True,
        #                     key_list    = [title],
        #                     dict_item   = uvalue_option[k2],
        #                     type_list   = v2,
        #                 )

        chk_vals_types(
            parent_name = 'uvalue_option',
            key_name    = 'custom_val_max',
            vals        = [uvalue_option['custom_val_max']],
            type_list   = [int],
        )
        if uvalue_option['custom_val_max'] < 0:
            error_msg = 'custom_val_max is more than 0.'
            raise ValueError(error_msg)
        elif uvalue_option['custom_val_max'] >= 1:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['custom_type'],
                dict_item   = uvalue_option,
                type_list   = [dict],
            )
            for type_name in list(uvalue_option['custom_type']):
                if (type_name in list(type_info)) is True:
                    chk_json_item(
                        parent_name = 'custom_type',
                        essential   = True,
                        key_list    = [type_name],
                        dict_item   = uvalue_option['custom_type'],
                        type_list   = [dict],
                    )
                    for k, v in type_info[type_name]['option'].items():
                        chk_json_item(
                            parent_name = type_name,
                            essential   = True,
                            key_list    = [k],
                            dict_item   = uvalue_option['custom_type'][type_name],
                            type_list   = v,
                        )
                else:
                    error_msg = str(uvalue_type) + ' is unknown type.'
                    raise ValueError(error_msg)

    elif uvalue_type == 'dict':
        my_type = type_info[uvalue_type]['my_type']

        chk_vals_types(
            parent_name = param_name,
            key_name    = 'default_value',
            vals        = [default_value],
            type_list   = [my_type, type(None)],
        )

        type_list = None
        if type(default_value) is type(None):
            type_list = [my_type]
        elif type(default_value) is not type(None):
            type_list = [my_type, type(None)]
        if type_list is not None:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['none_val'],
                dict_item   = uvalue_option,
                type_list   = type_list,
            )

        none_val = uvalue_option['none_val']

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

        dict_title = uvalue_option['item_titles']
        dict_type  = uvalue_option['value_types']

        title_key_list, title_layer_dict = search_dict_keys(
            dict_item  = dict_title,
            single_key = True,
            key_list   = [],
            layer_num  = 0,
        )
        title_layer_list = []
        [title_layer_list.extend(v) for v in title_layer_dict.values()]
        title_layer_list = list(set(title_layer_list))

        title_dict_len = len(title_key_list)
        key_list, layer_dict = search_dict_keys(
            dict_item  = dict_title,
            single_key = False,
            key_list   = [],
            layer_num  = 0,
        )

        if title_dict_len != len(key_list):
            error_msg = 'dict_title is single layer.'
            raise ValueError(error_msg)

        # for k in key_list:
        #     dict_title2chk = dict_title[k]
        #     chk_vals_types(
        #         parent_name = 'item_titles',
        #         key_name    = k,
        #         vals        = [dict_title2chk],
        #         type_list   = [eval(dict_type[k])],
        #     )

        dict_val_search = {
            'default_value' : default_value,
            'none_val'      : none_val,
        }
        for k, v in dict_val_search.items():
            if v is not None:
                key_list, layer_dict = search_dict_keys(
                    dict_item  = v,
                    single_key = False,
                    key_list   = [],
                    layer_num  = 0,
                )
                layer_list = []
                [layer_list.extend(v) for v in layer_dict.values()]
                layer_list = list(set(layer_list))
                if layer_list != [] and layer_list != title_layer_list:
                    error_msg = str(k) + ' has different layer number '\
                        'compared with dict_title. {0} vs {1}'.format(
                            layer_list,
                            title_layer_list
                        )
                    raise ValueError(error_msg)

                # title_list = list(layer_dict)
                # for k, v in type_info.items():
                #     for title in title_list:
                #         if dict_type[title] == k:
                #             for k2, v2 in v['option'].items():
                #                 chk_json_item(
                #                     parent_name = k2,
                #                     essential   = True,
                #                     key_list    = [title],
                #                     dict_item   = uvalue_option[k2],
                #                     type_list   = v2,
                                # )

        chk_vals_types(
            parent_name = 'uvalue_option',
            key_name    = 'custom_val_max',
            vals        = [uvalue_option['custom_val_max']],
            type_list   = [int],
        )
        if uvalue_option['custom_val_max'] < 0:
            error_msg = 'custom_val_max is more than 0.'
            raise ValueError(error_msg)
        elif uvalue_option['custom_val_max'] >= 1:
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = ['custom_type'],
                dict_item   = uvalue_option,
                type_list   = [dict],
            )
            for type_name in list(uvalue_option['custom_type']):
                if (type_name in list(type_info)) is True:
                    chk_json_item(
                        parent_name = 'custom_type',
                        essential   = True,
                        key_list    = [type_name],
                        dict_item   = uvalue_option['custom_type'],
                        type_list   = [dict],
                    )
                    for k, v in type_info[type_name]['option'].items():
                        chk_json_item(
                            parent_name = type_name,
                            essential   = True,
                            key_list    = [k],
                            dict_item   = uvalue_option['custom_type'][type_name],
                            type_list   = v,
                        )
                else:
                    error_msg = str(uvalue_type) + ' is unknown type.'
                    raise ValueError(error_msg)

    elif uvalue_type == 'None':
        my_type = type_info[uvalue_type]['my_type']

        for k, v in type_info[uvalue_type]['option'].items():
            chk_json_item(
                parent_name = 'uvalue_option',
                essential   = True,
                key_list    = [k],
                dict_item   = uvalue_option,
                type_list   = v,
            )

    else:
        error_msg = '\'' + str(uvalue_type) + '\' does not support.'
        raise ValueError(error_msg)


def chk_format_node_param_test():
    for node_name in all_node_name:
        node_dict = all_node[node_name]
        yield chk_format_node, node_name, node_dict

    for param_name in all_param_name:
        param_dict = all_param[param_name]
        yield chk_format_param, param_name, param_dict


if __name__ == '__main__':
    print(__doc__)
    for node_name in all_node_name:
        node_dict = all_node[node_name]
        chk_format_node( node_name, node_dict )

    for param_name in all_param_name:
        param_dict = all_param[param_name]
        chk_format_param( param_name, param_dict )




#!/usr/bin/env python3

CAM_MOD            = 57.0214637206
DIFFERENCE_14N_15N = 0.997035
PROTON             = 1.00727646677

NITROGENS = {
    'A' : 1,
    'C' : 1,
    'D' : 1,
    'E' : 1,
    'F' : 1,
    'G' : 1,
    'H' : 3,
    'I' : 1,
    'K' : 2,
    'L' : 1,
    'M' : 1,
    'N' : 2,
    'P' : 1,
    'Q' : 2,
    'R' : 4,
    'S' : 1,
    'T' : 1,
    'V' : 1,
    'W' : 2,
    'Y' : 1,
}

# mass of all amino acid difference between 15N and 14N
DICT_15N_DIFF = {
    'A' : 0.997035,
    'C' : 0.997035,
    'D' : 0.997035,
    'E' : 0.997035,
    'F' : 0.997035,
    'G' : 0.997035,
    'H' : 2.991105,
    'I' : 0.997035,
    'K' : 1.99407,
    'L' : 0.997035,
    'M' : 0.997035,
    'N' : 1.99407,
    'P' : 0.997035,
    'Q' : 1.99407,
    'R' : 3.98814,
    'S' : 0.997035,
    'T' : 0.997035,
    'V' : 0.997035,
    'W' : 1.99407,
    'Y' : 0.997035
}

# Dictionary containing all accepted file extensions in Ursgal
FILE_EXTENSIONS = {
    '.dat' : {
        'short_name': 'DAT',
        'long_name': 'DAT (Mascot)',
        'same_extension': [],
        'description': \
            'Mascot specific identification output format.'
            'Mixture of csv, xml and configuration style file.'
    },
    '.csv' : {
            'short_name'     : 'CSV',
            'long_name'      : 'CSV (Comma Delimited)',
            'same_extension' : [],
            'description'    : \
                'The comma-separated values (CSV) file format is a '\
                'tabular data format that has fields separated by the '\
                'comma character and quoted by the double quote '\
                'character.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.cfg' : {
            'short_name'     : 'CFG',
            'long_name'      : 'CFG (pGlyco)',
            'same_extension' : [],
            'description'    : \
                'pGlyco specific parameter file format.',
    },
    '.den' : {
        'short_name'     : 'DEN',
        'long_name'      : 'DEN (XMVB Density Data)',
        'same_extension' : [],
        'description'    : \
            'DEN file is a XMVB Density Data. Xiamen Valence Bond '\
            '(XMVB) is a quantum chemistry program for performing '\
            'electronic structure calculations based on the '\
            'non-orthogonal Valence Bond methods.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.dta' : {
        'short_name'     : 'DTA',
        'long_name'      : 'DTA (Stata Data)',
        'same_extension' : [],
        'description'    : \
            'DTA file is a Stata Data File. Stata is a '\
            'general-purpose statistical software package created in '\
            '1985 by StataCorp. It is used by many businesses and '\
            'academic institutions around the world.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.dta.txt' : {
        'short_name'     : 'DTA_Text',
        'long_name'      : 'DTA_Text (Text Fromat of DTA)',
        'same_extension' : [],
        'description'    : \
            'Text format of DTA format.',
    },
    '.fasta' : {
        'short_name'     : 'FASTA',
        'long_name'      : 'FASTA (Sequence Alignment)',
        'same_extension' : ['.fa', '.mpfa', '.fna', '.fsa', '.fas'],
        'description'    : \
            'FASTA file is a FASTA Sequence. In bioinformatics, FASTA '\
            'format is a text-based format for representing either '\
            'nucleotide sequences or peptide sequences, in which '\
            'nucleotides or amino acids are represented using '\
            'single-letter codes.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.gaml' : {
        'short_name'     : 'GAML',
        'long_name'      : 'GAML (GAML spectra)',
        'same_extension' : [],
        'description'    : \
            'GAML spectra',
    },
    '.idx.gz' : {
        'short_name'     : 'igzip',
        'long_name'      : 'igzipped mzML (MS data)',
        'same_extension' : [],
        'description'    : \
            'Compressed mzML, indexed gzip according to pymzML2.0',
    },
    '.kojak.txt' : {
        'short_name'     : 'Kojak',
        'long_name'      : 'Kojak (Kojak result)',
        'same_extension' : [],
        'description'    : \
            'Kojak result',
    },
    '.mgf' : {
        'short_name'     : 'MGF',
        'long_name'      : 'MGF (Mascot Generic Format)',
        'same_extension' : [],
        'description'    : \
            'The Mascot generic format for a data file submitted to Mascot',
    },
    '.ms2' : {
        'short_name'     : 'MS2',
        'long_name'      : 'MS2 (MS data)',
        'same_extension' : [],
        'description'    : \
            'Mass spectrometry data format',
    },
    '.mzData' : {
        'short_name'     : 'mzData',
        'long_name'      : 'mzData (MS data)',
        'same_extension' : [],
        'description'    : \
            'This format was deprecated, and was replaced by mzML',
    },
    '.mzid' : {
        'short_name'     : 'mzid',
        'long_name'      : 'mzid (MS data)',
        'same_extension' : [],
        'description'    : \
            'Mass spectrometry data format',
    },
    '.mzid.gz' : {
        'short_name'     : 'mzid.gz',
        'long_name'      : 'mzid.gz (Compressed mzid)',
        'same_extension' : [],
        'description'    : \
            'Compressed mzid',
    },
    '.mzML' : {
        'short_name'     : 'mzML',
        'long_name'      : 'mzML (MS data)',
        'same_extension' : [],
        'description'    : \
            'Mass spectrometry data format',
    },
    '.mzML.gz' : {
        'short_name'     : 'mzML',
        'long_name'      : 'mzML (MS data)',
        'same_extension' : [],
        'description'    : \
            'Compressed mzML',
    },
    '.mzXML' : {
        'short_name'     : 'mzXML',
        'long_name'      : 'mzXML (MS data)',
        'same_extension' : [],
        'description'    : \
            'This format was replaced by mzML',
    },
    '.pep.xml' : {
        'short_name'     : 'pep.xml',
        'long_name'      : 'pep.xml (Kojak result)',
        'same_extension' : [],
        'description'    : \
            'Kojak result',
    },
    '.perc.inter.txt' : {
        'short_name'     : 'perc.inter.txt',
        'long_name'      : 'perc.inter.txt (Kojak result)',
        'same_extension' : [],
        'description'    : \
            'Kojak result',
    },
    '.perc.intra.txt' : {
        'short_name'     : 'perc.intra.txt',
        'long_name'      : 'perc.intra.txt (Kojak result)',
        'same_extension' : [],
        'description'    : \
            'Kojak result',
    },
    '.perc.loop.txt' : {
        'short_name'     : 'perc.loop.txt',
        'long_name'      : 'perc.loop.txt (Kojak result)',
        'same_extension' : [],
        'description'    : \
            'Kojak result',
    },
    '.perc.single.txt' : {
        'short_name'     : 'perc.single.txt',
        'long_name'      : 'perc.single.txt (Kojak result)',
        'same_extension' : [],
        'description'    : \
            'Kojak result',
    },
    '.pkl' : {
        'short_name'     : 'PKL',
        'long_name'      : 'PKL (pickle)',
        'same_extension' : [],
        'description'    : \
            'PKL file is a file created by the Python pickle module',
    },
    '.raw' : {
        'short_name'     : 'RAW',
        'long_name'      : 'RAW (ThermoFisher RAW format)',
        'same_extension' : [],
        'description'    : \
            'ThermoFisher RAW format',
    },
    '.ssl' : {
        'short_name'     : 'SSL',
        'long_name'      : 'SSL (Spectrum Sequence List)',
        'same_extension' : [],
        'description'    : \
            'Generic tab-delimited text file format supported by BiblioSpec',
    },
    '.svg' : {
        'short_name'     : 'SVG',
        'long_name'      : 'SVG (Scalable Vector Graphic)',
        'same_extension' : [],
        'description'    : \
            'SVG file is a Scalable Vector Graphic. SVG is a language '\
            'for describing two-dimensional graphics and graphical '\
            'applications in XML.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.tsv' : {
        'short_name'     : 'TSV',
        'long_name'      : 'TSV (Tab-Separated Values document)',
        'same_extension' : [],
        'description'    : \
            'TSV is a Tab-Separated Values document. It is very '\
            'simple textual data format which allows tabular data to '\
            'be exhanged between applications that use different '\
            'internal data formats.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.txt' : {
        'short_name'     : 'Text',
        'long_name'      : 'Text (General text format)',
        'same_extension' : ['.text'],
        'description'    : \
            'TXT file is a plain text. Plain text is textual '\
            'material, usually in a disk file, that is (largely) '\
            'unformatted.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.xml' : {
        'short_name'     : 'XML',
        'long_name'      : 'XML (Extensible Markup Language document)',
        'same_extension' : [],
        'description'    : \
            'XML file is an Extensible Markup Language document. XML '\
            'is a simple, very flexible text format derived from SGML '\
            '(ISO 8879). Originally designed to meet the challenges '\
            'of large-scale electronic publishing, XML is also '\
            'playing an increasingly important role in the exchange '\
            'of a wide variety of data on the Web and elsewhere.\n'\
            '(DataTypes.net; https://datatypes.net)',
    },
    '.xml.gz' : {
        'short_name'     : 'XML.gz',
        'long_name'      : 'XML.gz (Compressed mzid)',
        'same_extension' : [],
        'description'    : \
            'Compressed xml',
    }
}

ENGINE_TYPES = {
    'converter'                     : 'convert',
    'controller'                    : None,
    'cross_link_search_engine'      : 'search_mgf',
    'de_novo_search_engine'         : 'search_mgf',
    'fetcher'                       : 'fetch_file',
    'misc_engine'                   : 'execute_misc_engine',
    'meta_engine'                   : 'combine_search_results',
    'protein_database_search_engine': 'search_mgf',
    'quantification_engine'         : 'quantify',
    'spectral_library_search_engine': 'search_mgf',
    '_test'                         : 'execute_misc_engine',
    'validation_engine'             : 'validate',
    'visualizer'                    : 'visualize',
}

UCONTROLLER_FUNCTIONS = {}
for engine_type, ucontoller_function in ENGINE_TYPES.items():
    if ucontoller_function is None:
        continue
    if ucontoller_function not in UCONTROLLER_FUNCTIONS.keys():
        UCONTROLLER_FUNCTIONS[ucontoller_function] = []
    UCONTROLLER_FUNCTIONS[ucontoller_function].append(engine_type)


#!/usr/bin/env python3
# encoding: utf-8
'''

Test the unify_csv function for omssa engine

'''
import ursgal
import csv
import pickle
import os

uc = ursgal.UController(
    profile = 'LTQ XL low res',
    force  = False
)

upeptide_mapper_main = uc.unodes['upeptide_mapper_1_0_0']['class'].import_engine_as_python_function(
    'main'
)

input_csv = os.path.join(
    'tests',
    'data',
    'test_upeptide_mapper.csv'
)
output_csv = os.path.join(
    'tests',
    'data',
    'test_upeptide_mapper_pmap.csv'
)

params = {
    'translations' : {
        'modifications' : [
            'M,opt,any,Oxidation',        # Met oxidation
            'C,fix,any,Carbamidomethyl',  # Carbamidomethylation
            '*,opt,Prot-N-term,Acetyl',    # N-Acteylation[]
        ],
        'aa_exception_dict' : {
            'J' : {
                'original_aa' : ['I', 'L'],
            },
            'O' : {
                'original_aa' : ['K'],
                'unimod_name' : 'Methylpyrroline',
            },
        },
        'protein_delimiter'        : '<|>',
        'decoy_tag'                : 'decoy_',
        'word_len'                 : 6,
        'database' : os.path.join(
            'tests',
            'data',
            'BSA.fasta'
        ),
    },
    'prefix' : None
}

all_mapped_peptides = set()
ident_list = [ ]
for peptide_mapper_class_version in [ 'UPeptideMapper_v3', 'UPeptideMapper_v4' ]:
    params['translations']['peptide_mapper_class_version'] = peptide_mapper_class_version
    upeptide_mapper_main(
        input_file     = input_csv,
        output_file    = output_csv,
        params         = params,
    )

    for line_dict in csv.DictReader(open(output_csv, 'r')):
        ident_list.append( line_dict )
        all_mapped_peptides.add(line_dict['Sequence'])

def upeptide_mapper_test():
    for test_id, test_dict in enumerate(ident_list):
        yield upeptide_mapper, test_dict


def upeptide_mapper( test_dict ):
    for key in [
            'Sequence',
            'Modifications',
            'Is decoy',
            'Protein ID',
            'Sequence Start',
            'Sequence Stop',
            'Sequence Pre AA',
            'Sequence Post AA',
        ]:
        test_value = test_dict[key]
        expected_value = test_dict[ 'Expected {0}'.format(key) ]
        assert test_value == expected_value, '''
Unexpected value in column "{0}":
    test value:     {1}
    expected value: {2}
            '''.format(key, test_value, expected_value)

def non_map_test():
    ''' 
    LERAA would be an overlap of ProtD and ProtE and should not be mapped.
    '''
    assert 'LERAAA' not in all_mapped_peptides

if __name__ == '__main__':
    print(__doc__)
    for test_id, test_dict in enumerate(ident_list):
        upeptide_mapper(test_dict)



#!/usr/bin/env python
import ursgal
import importlib
import os
import sys
import pickle
import shutil


class csv2counted_results_1_0_0(ursgal.UNode):
    """csv2counted_results_1_0_0 UNode"""

    META_INFO = {
        'edit_version': 1.00,
        'name': 'Summarize Results as Tables',
        'version': '1.0.0',
        'release_date': None,
        'engine_type': {
            'converter': True
        },
        'output_extensions': ['.csv'],
        'input_extensions': ['.csv'],
        'output_suffix': 'counted',
        'in_development': False,
        'utranslation_style': 'csv2counted_results_style_1',
        'include_in_git': True,
        'distributable': True,
        'engine': {
            'platform_independent': {
                'arc_independent': {
                    'exe': 'csv2counted_results_1_0_0.py',
                },
            },
        },
        'citation': '',
    }

    def __init__(self, *args, **kwargs):
        super(csv2counted_results_1_0_0, self).__init__(*args, **kwargs)

    def _execute(self):
        '''
        Results (.csv) are summarized as table (.csv) containing all identified
        proteins, peptides, or other specified identifiers. For each sample,
        the peptide or spectral count for each identifier is given.

        Input file has to be a .csv

        Creates a _counted.csv file and returns its path.

        Columns containing the elements that should be counted (identifiers)
        are given as a list of headers using uc.params["identifier_column_names"].
        Columns defining a unique countable element (e.g. "Sequence", "Spectrum ID")
        are given as a list of headers using uc.params["count_column_names"].

        This can be used to create a SFINX (http://sfinx.ugent.be/) input file, 
        using: 
            uc.params["convert_to_sfinx"]=True
            uc.params["identifier_colum_names"]=["Protein ID"]
            uc.params["count_column_names"]=["Sequence"]

        '''

        print('[ -ENGINE- ] Executing conversion ..')
        self.time_point(tag='execution')
        csv2count_main = self.import_engine_as_python_function()
        if self.params['input_file'].lower().endswith('.csv') is False:
            raise ValueError('Trying to convert a non-csv file')

        output_file = os.path.join(
            self.params['output_dir_path'],
            self.params['output_file']
        )
        input_file = os.path.join(
            self.params['input_dir_path'],
            self.params['input_file']
        )

        csv2count_main(
            input_file=input_file,
            output_file=output_file,
            identifier_colum_names=self.params['translations']['identifier_column_names'],
            count_column_names=self.params['translations']['count_column_names'],
            count_by_file=self.params['translations']['count_by_file'],
            convert2sfinx=self.params['translations']['convert_to_sfinx'],
            keep_column_names=self.params['translations']['keep_column_names'],
        )

        self.print_execution_time(tag='execution')
        return output_file



#!/usr/bin/env python3
'''

Tests the reformat peptide function in ucore


'''
import re
import ursgal.ucore

TESTS = [
    {
        'unformated_peptide':   '(ac)SSSLM(ox)RPGPSR',
        'regex_list':[('\(ac\)','Acetyl'), ('\(ox\)','Oxidation') ],
        'result': 'SSSLMRPGPSR#Acetyl:0;Oxidation:5'
    },
    {
        'unformated_peptide':   '(ac)1234(ox)56(ox)7',
        'regex_list':[('\(ac\)','Acetyl'), ('\(ox\)','Oxidation') ],
        'result': '1234567#Acetyl:0;Oxidation:4;Oxidation:6'
    },
    {
        'unformated_peptide':   '(ac)1234(ox)56(ox)7',
        'regex_list':[('\(ox\)','Oxidation'),('\(ac\)','Acetyl')],
        'result': '1234567#Acetyl:0;Oxidation:4;Oxidation:6'
    },
    {
        'unformated_peptide':   '(ox)1234(ac)56(ox)7',
        'regex_list':[('\(ox\)','Oxidation'),('\(ac\)','Acetyl')],
        'result': '1234567#Oxidation:0;Acetyl:4;Oxidation:6'
    },
    {
        'unformated_peptide':   '(ac)1234(ox)56(ox)7',
        'regex_list':[ ('\(ox\)','Oxidation') ],
        'result': '(ac)1234567#Oxidation:8;Oxidation:10'
    },
    {
        'unformated_peptide':   '1234567',
        'regex_list':[ ('\(ox\)','Oxidation') ],
      'result':   '1234567'
    },
    {
        'unformated_peptide':   '1234567#',
        'regex_list':[ ('\(ox\)','Oxidation') ],
      'result':   '1234567'
    },
    {
        'unformated_peptide':   'AIVDC(cm)GFEHPSEVQHEC(cm)IPQAILGM(ox)DVLC(cm)QAK',
        'regex_list':[ ('\(ox\)','Oxidation'), ('\(cm\)','Carbamidomethyl') ],
        'result': 'AIVDCGFEHPSEVQHECIPQAILGMDVLCQAK#Carbamidomethyl:5;Carbamidomethyl:17;Oxidation:25;Carbamidomethyl:29'
    },
    {
        'unformated_peptide':   '12(cm)34(cm)5(ox)6(cm)78',
        'regex_list':[ ('\(ox\)','Oxidation'), ('\(cm\)','Carbamidomethyl') ],
        'result': '12345678#Carbamidomethyl:2;Carbamidomethyl:4;Oxidation:5;Carbamidomethyl:6'
    },
    {
        'unformated_peptide':   '12(cm)34(cm)5(ox)6(cm)78',
       'regex_list': [  ('\(cm\)','Carbamidomethyl'), ('\(ox\)','Oxidation') ],
       'result':  '12345678#Carbamidomethyl:2;Carbamidomethyl:4;Oxidation:5;Carbamidomethyl:6'
    },
    {
       'unformated_peptide':   'M(O)C(Cam)C(Cam)HDWRR',
       'regex_list': [  ('\(Cam\)','Carbamidomethyl'), ('\(O\)','Oxidation') ],
       'result':  'MCCHDWRR#Oxidation:1;Carbamidomethyl:2;Carbamidomethyl:3'
    },
    {
       'unformated_peptide':   '[Acetyl]C(Cam)M(O)M(O)MGASDGK',
       'regex_list': [  ('\(Cam\)','Carbamidomethyl'), ('\(O\)','Oxidation'), ('\[Acetyl\]','Acetyl') ],
       'result':  'CMMMGASDGK#Acetyl:0;Carbamidomethyl:1;Oxidation:2;Oxidation:3'
    },
    {
       'unformated_peptide':   ' SCCPCCM(O)ALHAGC(Carboxymethyl)R',
       'regex_list': [  ('\(Carboxymethyl\)','Carboxymethyl'), ('\(O\)','Oxidation') ],
       'result':  'SCCPCCMALHAGCR#Oxidation:7;Carboxymethyl:13'
    }
]


def handle_entry_test():
    for test_no, test_dict in enumerate(TESTS):
        yield _reformat_peptide, \
            test_no, \
            test_dict


def _reformat_peptide( test_no, test_dict ):
    formated_peptide = test_dict['unformated_peptide']
    for regex_pattern, unimod_name in test_dict['regex_list']:
        formated_peptide = ursgal.ucore.reformat_peptide(
            regex_pattern,
            unimod_name,
            formated_peptide
        )
    print( formated_peptide , test_dict['result'])
    assert formated_peptide == test_dict['result']

if __name__ == '__main__':
    print(__doc__)
    for test_no, test_dict in enumerate(TESTS):
        formated_peptide = test_dict['unformated_peptide']
        for regex_pattern, unimod_name in test_dict['regex_list']:
            formated_peptide = ursgal.ucore.reformat_peptide(
                regex_pattern,
                unimod_name,
                formated_peptide
            )
        print( formated_peptide , test_dict['result'])
        assert formated_peptide == test_dict['result']

